#!/usr/bin/perl
#=========================================================================
# Import CSV into Addressbook Utility
# Version 1.03
# 
# Copyright (C) 2006 Greg J. Zartman <greg@leiinc.com>
#
#------------------------------------------------------------------------
# CHANGELOG:
#
# Version 1.03 (December 19, 2006)
# - Bug fixes 
# - Update documentation
#
# Version 1.02 (December 16, 2006)
# - Allow for user defined csv delimiter
# - Allow for user defined csv column order
# - Provide 5 user defined import data fields
# - Clean up code and document throughout.
#
# Version 1.01 (December 13, 2006)
# - Bug fixes and continued development
#
# Version 1.00 (December 10, 2006)
# - Initial development release
#
#
#------------------------------------------------------------------------
# DOCUMENTATION:
#
# This utility imports delimited ascii (CSV) address/contact data into 
# the Addressbook application:  
# http://sourceforge.net/projects/theaddressbook/
#
# Syntax:  perl import.pl some_csv_text_file
#
# Data fields that can be imported using this utility:
# -firstname
# -middlename
# -lastname
# -addressline1
# -addressline2
# -city
# -state
# -zip
# -country
# -phone
# -fax
# 
# The only field that is really requred is the lastname field since
# Addressbook indexes contacts by lastname, however even this is
# not strictly enforced by the utility.  If lastname is missing
# the utility will store the contact under the lastname 
# unknown concatenated with the ID number for the contact
# (e.g. unknown224).  
#
# The utility also allows for 5 user defined contact fields.
# These fields will be stored in the addtionaldata table. See
# configuration options below for more details.
# 
# This utility does next to no checking of the data, therefore it is
# up to the user to make sure the data is clean before attempting an
# import. To aide the user in making sure the data is ready for import
# into Addressbook, a testmode is provided.  While in testmode, the 
# utility will process csv data, but not save the data to the 
# Addressbook database tables.  Processed data is sent to STDOUT
# for examination and troubleshooting.  
#
# TO INITIATE TESTMODE:
# 
# Syntax: perl import.pl some_csv_text_file test
#
# It will be difficult to examine the data as it scrolls by in 
# STDOUT, so it may be useful to redirect STDOUT to a file:
# 
# perl import.pl some_csv_text_file test > output.txt
#
# Following this documentation section are a set of configuration
# parameters to configuration the utility.  Documentation for the
# parameters is included in notes with each parameter.  It is 
# recommended that you read these throughly and perform at least 
# one run of the utility in testmode prior to preforming a live import.
#
#------------------------------------------------------------------------
# BUGS:
#
# Please send bug reports to greg@leiinc.com.  I will work to 
# fix them and post an updated version of this utility as quickly
# as possible
#
#------------------------------------------------------------------------
# LICENSE: 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License or more details.
#=======================================================================

use strict;
use warnings;
use DBI;

#=======================================================================
#
#                  CONFIGURATION PARAMETERS 
#
#=======================================================================


#-----------------------------------------------------------------------
# MYSQL Database Variables
# 
# Specify the following variables to gain access to the addressbook 
# database.  
#-----------------------------------------------------------------------

my $dbname   = 'addressbook'; #database name
my $username = 'addressbook'; #usename to gain access to the database
my $password = ''; #password to gain access to the database


#-----------------------------------------------------------------------
# Import Address Type
#
# Specify what type of addresses you are importing 
# (e.g., Business, Personal).  
# If this field is not important to you, just leave it blank
#-----------------------------------------------------------------------

my $address_type = 'Business';

#-----------------------------------------------------------------------
# Database Keys
#
# This option is used to tell the utility if the csv import data 
# contains unique keys or not as part of the import features.  
# A database key is a unique integer for each contact that the database
# uses to keep track of data.
# All databases use unique keys to keep track of data.  This import
# options gives you the ability to control how this key is assigned to 
# the data you are importing.  Set this value to 1 to specify that csv 
# import data contains unique keys.  Keys must be integers and must
# be unique.  The utility will check the import keys to make sure 
# they are integers and unique. Leave this value blank or set it to 0 
# to let the import utility designate keys for you.
#-----------------------------------------------------------------------

my $keyed_data = 1; 

#-----------------------------------------------------------------------
# CSV Delimiter
#
# Designate the type of delimiter used to delimit data in the CSV
# import file.  The default delimiter is a comma delimiter, meaning 
# that if you leave this field blank, the utility will assume the
# data is comma delimited.  Please note that this parameter will only
# accept keyboard characters as delimiters.  If you wish to designate
# tab delimited, for example, you will need to define a perl regex
# below for the tab character.  Future versions of this utility may 
# include several pre-defined delimiter options to choose from. 
# After some testing with importing data, I am finding that using a 
# pipe is about the best delimiter for contact data.  Using the comma
# delimiter can somethings run you into problems for certain types of 
# data (e.g. Some Company Name, Inc.)
#-----------------------------------------------------------------------

my $csv_delimiter = '|';

#-----------------------------------------------------------------------
# CSV Column Order
#
# Use this variable to define the order of the CSV data.  If you don't 
# need to import a certain column of data, set that field to 0.
#-----------------------------------------------------------------------
my %csv =(
          id         => 1,
          firstname  => 2,
          middlename => 3,
          lastname   => 4,
          line1      => 5,
          line2      => 6,
          city       => 7,
          state      => 8,
          zip        => 9,
          country    => 10,
          phone1     => 11,
          phone2     => 12,
          option1    => 13,
          option2    => 14,
          option3    => 15,
          option4    => 16,
          option5    => 17 
         );

#-----------------------------------------------------------------------
# User Defined Data Items
#
# Specify the names of the user defined data fields that you'd like to 
# import.  The utility will ignore those fields that are left blank.
#
#-----------------------------------------------------------------------

my %options =(
               1 => 'title',
               2 => 'company',
               3 => 'job_title',
               4 => '',
               5 => ''
             );

#-----------------------------------------------------------------------
# Lastname substitute optional data item
#
# Specify the optional data item that should be used when
# the contact lastname is not present.  This is useful if you
# are importing contacts with company names, but no specific
# contact person.  The optional data field will be used in place
# of the lastname field, as Addressbook sorts contacts by last name.
#
# Valid entries for this var include: option1, option2, option3, 
# option4, or option5.
# 
# If this var is not important to you, leave it blank.  This utility
# will default any contacts with a missing last name to unknown if this 
# var is left blank or an incorrect value is specified.
#-----------------------------------------------------------------------
my $ln_substitute = 'option2';


########################################################################
########################################################################
##
## DO NOT ALTER ANYTHING BEYOND THIS POINT UNLESS YOU ARE PERL SAVY
##
########################################################################
########################################################################

#-----------------------------------------------------------------------
# Designate initial vars
#-----------------------------------------------------------------------

## Set CSV delimiter for the import file.  Default to comma unless
## specified in config area above. Also, we need to escape out 
## keyboard characters.  This may cause problems down the line 
## if we try to use a tab regex, but we'll deal with that when
## it actually happens
if ($csv_delimiter) {
    $csv_delimiter = '\\' . $csv_delimiter;
}
else {
    $csv_delimiter = ',' unless ($csv_delimiter);
}

## Initialize dbase handles since we are using them for the first time
## in a subroutine and we want them scoped for the main routine.
my ($dbh,$sth1,$sth2,$sth3);

## Intialize dbase key var
my $id;

#-----------------------------------------------------------------------
# Perform data pre-processing and configuration checks
#-----------------------------------------------------------------------

print "CSV to Addressbook started...\n\n";
print "Checking configuration...\n";

## Pull command line arguments
my $filename   = $ARGV [0];  #path and file name to csv
my $run_option = $ARGV [1]; #utility run option.

## Check to make sure a file name was specified.  If not, bail
unless ($filename) {
    print "\n\n Bad command syntax.  Unable to start utility...\n";
    print "Syntax: import.pl csvfilename\n";
    exit;
}

##Send what type of keys we are importing to stdout
if ($keyed_data) {
    print "Data keys included with import data...\n";
}
else {
    print "Data keys not included with import data...\n";
}

## Process run options
my $testmode=0;
$run_option = lc($run_option);
if ($run_option eq 'test') {
    $testmode = 1;
    print "Test mode specified.  No data will be imported...\n";
}

## Run a quick check to make sure keys are either defined or we are 
## pulling them from the dbase. 
if ($keyed_data && !$csv{'id'}) {
    print "Unable to determine data keys.  Either include data \n";
    print "keys in your export file or set the utility to \n";
    print "determine them for you.  No data imported....\n";
    exit;
}

## Open the csv text file and load data into memory
print "Opening CSV import file $filename...\n";
open(FILE, "<", $filename)or die "Couldn't open $filename: $!";
my @textfile = <FILE>;
close FILE;

## Test then open a connection to the database
test_dbase_connection();
open_dbase_connection() unless $testmode;

## Test import keys if data is keyed
test_dbase_keys() if ($keyed_data);

## Determine the starting dbase key to use if import data is not keyed
my $starting_key = get_starting_dbase_key() unless $keyed_data;

print "Configuration is good.  Processing data...\n\n";

#-----------------------------------------------------------------------
# Perform Data Import
#-----------------------------------------------------------------------

## Cycle through the csv data and import into MYSQL.  As data
## is imported into MYSQL, each data element will be sent
## to STDOUT after it is imported.  If testmode is designated, 
## the data will be processed, sent to STDOUT, but not imported
## into MYSQL.
my $count = 0;
my @fields=();
my ($firstname,$middlename,$lastname,$line1,$line2,$city,$state,$zip,
    $country,$phone1,$phone2,$option1,$option2,$option3,$option4,
    $option5); 

$id = $starting_key unless($keyed_data);

foreach (@textfile) {

    #Remove newline character
    chomp $_;

    #Skip blank entries
    next unless ($_);

    # Split data by delimiter and store each field in a var
    @fields = split ($csv_delimiter, $_); 

    #my $counter=1;
    #my $field;
    #foreach $field (@fields) {
    #    print "$counter $field\n";
    #    $counter++;
    #}

    # Save import fields to human readable vars so we can keep 
    # track of what is going on.  
    if ($keyed_data){
        if ($fields[$csv{'id'}-1]) {
            $id = $fields[$csv{'id'}-1];
        }
        else {$id = '';}
    }
    if ($csv{'firstname'} && $fields[$csv{'firstname'}-1]){
        $firstname = $fields[$csv{'firstname'}-1];
    }
    else {$firstname = '';}
    if ($csv{'middlename'} && $fields[$csv{'middlename'}-1]){
        $middlename = $fields[$csv{'middlename'}-1];
    }
    else {$middlename ='';}
    if ($csv{'lastname'} && $fields[$csv{'lastname'}-1]) {
        $lastname   = $fields[$csv{'lastname'}-1];
    }
    else {$lastname='';}
    if ($csv{'line1'} && $fields[$csv{'line1'}-1]) {
        $line1 = $fields[$csv{'line1'}-1];
    }
    else {$line1 = '';}
    if ($csv{'line2'} && $fields[$csv{'line2'}-1]) {
        $line2 = $fields[$csv{'line2'}-1];
    }
    else {$line2 = '';}
    if ($csv{'city'} && $fields[$csv{'city'}-1]) {
        $city = $fields[$csv{'city'}-1];
    }
    else {$city = '';}
    if ($csv{'state'} && $fields[$csv{'state'}-1]) {
        $state = $fields[$csv{'state'}-1];
    }
    else {$state = '';}
    if ($csv{'zip'} && $fields[$csv{'zip'}-1]) {
        $zip = $fields[$csv{'zip'}-1];
    }
    else {$zip = '';}

    if ($csv{'country'} && $fields[$csv{'country'}-1]) {
        $country = $fields[$csv{'country'}-1];
    }
    else {$country = '';}
    if ($csv{'phone1'} && $fields[$csv{'phone1'}-1]) {
        $phone1 = $fields[$csv{'phone1'}-1];
    }
    else {$phone1 = '';}
    if ($csv{'phone2'} && $fields[$csv{'phone2'}-1]) {
        $phone2 = $fields[$csv{'phone2'}-1];
    }
    else {$phone2 = '';}

    if ($csv{'option1'} && $fields[$csv{'option1'}-1]) {
        $option1 = $fields[$csv{'option1'}-1];
    }
    else {$option1 = '';}

    if ($csv{'option2'} && $fields[$csv{'option2'}-1]) {
        $option2 = $fields[$csv{'option2'}-1];
    }
    else {$option2 = '';}
    if ($csv{'option3'} && $fields[$csv{'option3'}-1]) {
        $option3 = $fields[$csv{'option3'}-1];
    }
    else {$option3 = '';}
    if ($csv{'option4'} && $fields[$csv{'option4'}-1]) {
        $option4 = $fields[$csv{'option4'}-1];
    }
    else {$option4 = '';}
    if ($csv{'option5'} && $fields[$csv{'option5'}-1]) {
        $option5 = $fields[$csv{'option5'}-1];
    }
    else {$option5 = '';}

    #set address id = contact id.  
    my $refid = $id;

    #Check to see if entry has a lastname.  If not, then
    #set it to the ln_substitute option or unknown if this var is 
    #invalid or not present
    $ln_substitute = lc($ln_substitute);
    
    unless($lastname) {
        if ($fields[$csv{$ln_substitute}-1]) {
            $lastname = $fields[$csv{$ln_substitute}-1];
        }
        else {$lastname = 'unknown' . $id;}
    }   
    
    # Save data to MySQL
    unless ($testmode) {

        # Save data to address table
        $sth1->execute($refid, $id, $address_type, $line1, $line2, 
        $city, $state, $zip, $country, $phone1, $phone2)
        or die "Error executing SQL query: " . $sth1->errstr;

        # Save data to contact table
        $sth2->execute($id, $firstname, $lastname, $middlename, $refid)
         or die "Error executing SQL query: " . $sth2->errstr;

        # Save optiona1 data item t0 additionaldata table
        if ($options{1}) {
            $sth3->execute($id, $options{1}, $option1)
             or die "Error executing SQL query: " . $sth3->errstr;
        }
        
        # Save option2 data item to additionaldata table
        if ($options{2}) {
            $sth3->execute($id, $options{2}, $option2)
             or die "Error executing SQL query: " . $sth3->errstr;
        }

        # Save option3 data item to to additionaldata table
        if ($options{3}) {
            $sth3->execute($id, $options{3}, $option3)
             or die "Error executing SQL query: " . $sth3->errstr;
        }
        
        # Save option4 data item to to additionaldata table
        if ($options{4}) {
            $sth3->execute($id, $options{4}, $option4)
             or die "Error executing SQL query: " . $sth3->errstr;
        }
        
        # Save option5 data item to to additionaldata table
        if ($options{5}) {
            $sth3->execute($id, $options{5}, $option5)
             or die "Error executing SQL query: " . $sth3->errstr;
        }

    }

    
    # Send current data element to stdout
    send_to_stdout();

    $id++ unless ($keyed_data);
    $count++;

}

print "Finished, $count fields imported!!!\n";


#=======================================================================
# SUBROUTINES
#=======================================================================

#-----------------------------------------------------------------------
#  Test Database Connection Subroutine
#  -This routine simply tests to see if we can connect to the database
#-----------------------------------------------------------------------
sub test_dbase_connection {

print "Testing dbase connection...\n";
$dbh = DBI->connect("DBI:mysql:$dbname", $username, $password)
       or warn "Couldn't connect to db $dbname:" . DBI->errstr;

my $dc = $dbh->disconnect;

return (1);
}


#-----------------------------------------------------------------------
# Open Database Connection Subroutine
#  -This routine opens a connection to mysql and defines the insert 
#   query strings for uploading our data to SQL.
#-----------------------------------------------------------------------
sub open_dbase_connection {

print "Opening a connection to the database...\n";

$dbh = DBI->connect("DBI:mysql:$dbname", $username, $password)
       or die "Couldn't connect to db $dbname:" . DBI->errstr;

## Define a query string to import data into the address table
$sth1 = $dbh->prepare("INSERT INTO
        address (refid, id, type, line1, line2, city, state, zip,
        country, phone1, phone2)
        values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)")
        or die "Couldn't create statement: " . $dbh->errstr;

## Define a query string to import data into contact table
$sth2 = $dbh->prepare("INSERT INTO contact (
        id, firstname, lastname, middlename, primaryAddress)
        values (?, ?, ?, ?, ?)")
        or die "Couldn't create statement: " . $dbh->errstr;

## Define a query string to import data into additional data table
$sth3 = $dbh->prepare("INSERT INTO additionaldata (id, type, value)
        values (?, ?, ?)")
        or die "Couldn't create statement: " . $dbh->errstr;

return(1);
}

#-----------------------------------------------------------------------
# Test Import Dbase Keys
#  -This routine tests import database keys if they are included
#   in the CSV import file.  
#-----------------------------------------------------------------------
sub test_dbase_keys {
my %unique_keys = ();
my @the_rest = ();
my $last_valid_key = 0;

print "Testing import keys...\n";

foreach (@textfile) {
   
    # remove the end of line character
    chomp $_;
    
    # Skip blank entries
    next unless ($_); 

    # Split data by comma delimiter and store each field in a var
    my $delimeter = '\\'.$csv_delimiter;
    ($id,@the_rest) = split($csv_delimiter, $_);
    
    # Make sure data key is an integer.  If it isn't, bail out
    # with an error message.
    unless($id =~ /^[+-]?\d+$/) {
        print "Non-Integer data key found.  Review your import data and \n";
        print "sure that every key is an integer.\n";
        print "Current field key:  $id\n";
        print "Last valid data key is $last_valid_key.\n";
        print "No data imported.  Exiting....\n";
        exit;
    }

    # Make sure data key is unique.  If it isn't, bail out with an
    # error messagbe
    if ($unique_keys{$id}) {
       print "Duplicate data key found.  Review your import data and make\n";
       print "sure that every entry has a unique key.\n";
       print "Duplicate key:  $id\n";
       print "Key for last valid entry:  $last_valid_key\n";
       print "No data imported.  Exiting....\n";
       exit;
    }

   # Record key for last valid entry
   $last_valid_key = $id;

   # Register key
   $unique_keys{$id} = 1;

}

return(1);

}

#-----------------------------------------------------------------------
# Determine Starting Dbase Key
#  -Query the database and return the highest key.  Our starting key
#   will be made one higher than this number.
#-----------------------------------------------------------------------
sub get_starting_dbase_key {

my $query_key;

my $sth4 = $dbh->prepare("SELECT MAX(id) from contact")
 or die "Couldn't query id key from contact table: " . $dbh->errstr;

$sth4 -> execute() or 
 die "Couldn't query id key from contact table: " . $sth4->errstr;

$query_key = $sth4->fetchrow;

return (1) unless ($query_key);


return($query_key+1);

}


#-----------------------------------------------------------------------
# Send Data Element to STDOUT
#  -This routine simply prints the current process data element to STDOUT.
#-----------------------------------------------------------------------
sub send_to_stdout {

print "Processing entry number $count \n";
print "Database key: $id \n";
print "First Name: $firstname \n";
print "Middle Name: $middlename \n";
print "Last Name: $lastname \n";
print "Address Line1:  $line1 \n";
print "Address Line2:  $line2 \n";
print "City: $city \n";
print "State:  $state \n";
print "Zip:  $zip \n";
print "Country: $country \n";
print "Phone:  $phone1\n";
print "Fax:  $phone2 \n";
print "Option1($options{1}): $option1\n";
print "Option2($options{2}): $option2\n";
print "Option3($options{3}): $option3\n";
print "Option4($options{4}): $option4\n";
print "Option5($options{5}): $option5\n";
print "\n";

return(1);
}
